/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "SArrayIndex.h"
#include "SArrayBasedFindTask.h"

namespace GB2 {

#define PCHAR_MATCH(x, y) (*(x) == *(y) && *(x) != unknownChar)


SArrayBasedFindTask::SArrayBasedFindTask(  SArrayIndex* i, const SArrayBasedSearchSettings& s )
: Task("SArrayBasedFindTask", TaskFlag_None), index(i), config(s)
{
    assert(index);
}

void SArrayBasedFindTask::run()
{
     runSearchWithMismatches();
}

void SArrayBasedFindTask::runSearch()
{
    bool haveResults = true;
    SArrayIndex::SAISearchContext context;
    const char* querySeq = config.query.constData();
    if (config.useBitMask) {
        const quint32* bm = config.bitMask;
        int charBitsNum = config.bitMaskCharBitsNum;
        quint32 bitValue = 0; 
        int wCharsInMask = index->getCharsInMask();
        const char* posS = querySeq;
        char unknownChar = config.unknownChar;
        for (int cleanChars = 0; cleanChars < wCharsInMask; posS++) {
            if (*posS == unknownChar) {
                cleanChars = 0;
                bitValue = 0;
            } else {
                bitValue = (bitValue << charBitsNum) | bm[uchar(*posS)];
                cleanChars++;
            }
        }
        haveResults = index->findBit(&context,bitValue,querySeq);
    } else {
        haveResults = index->find(&context,querySeq);
    }

    if (haveResults) {
        int pos = -1;
        while( ( pos = index->nextArrSeqPos(&context) ) != -1 ) {
            results.append(pos + 1);
        }
    }
   
}

void SArrayBasedFindTask::runSearchWithMismatches()
{
    const char* querySeq = config.query.constData();
    const char* arraySeq = index->getIndexedSequence();
    char unknownChar = config.unknownChar;
    int CMAX = config.nMismatches;
    int W = config.query.size();
    int q = W / (CMAX + 1);
    int prefixSize = index->getPrefixSize();
    
    assert(prefixSize <=  q);
    if (prefixSize > q) {
        setError( QString("Too large SArrayIndex window (%1) for %2-mismatch search").arg(prefixSize).arg(CMAX) );
        return;
    }
    
    for (int i = 0; i < W - prefixSize + 1; ++i) {
        const char* seq = querySeq + i;
        SArrayIndex::SAISearchContext context;
        bool haveResults = false;
        if (config.useBitMask) {
            const quint32* bm = config.bitMask;
            int charBitsNum = config.bitMaskCharBitsNum;
            quint32 bitValue = 0; 
            int wCharsInMask = index->getCharsInMask();
            const char* posS = querySeq;
            char unknownChar = config.unknownChar;
            for (int cleanChars = 0; cleanChars < wCharsInMask; posS++) {
                if (*posS == unknownChar) {
                    cleanChars = 0;
                    bitValue = 0;
                } else {
                    bitValue = (bitValue << charBitsNum) | bm[uchar(*posS)];
                    cleanChars++;
                }
            }
            haveResults = index->findBit(&context,bitValue,seq);
        } else {
            haveResults = index->find(&context,seq);
        }
        if (!haveResults) {
            continue;
        }
        int pos = -1;
        const char* endS = querySeq + W;
        const char* endA = arraySeq + index->getSequenceLength();
        while( ( pos = index->nextArrSeqPos(&context) ) != -1 ) {
            int c = 0;
            // forward collect 
            const char* posS = seq + prefixSize;
            const char* posA = arraySeq + pos + prefixSize;
            for ( ; (posS < endS) && (c <= CMAX); posS++, posA++) {
                if (posA >= endA) {
                    // out of arrraySeq boundaries - > do not need to continue
                    c = CMAX + 1;
                    break;
                }
                c += PCHAR_MATCH(posS, posA) ? 0 : 1; 
            }
            
            // backward collect
            posS = seq - 1;
            posA = arraySeq + pos - 1;
            
            for ( ; (posS >= querySeq) && (c <= CMAX); posS--, posA--) {
                if (posA < arraySeq) {
                    // out of arrraySeq boundaries - > do not need to continue
                    c = CMAX + 1;
                    break;
                }
                c += PCHAR_MATCH(posS, posA) ? 0 : 1; 
            }
            int result = pos - i + 1;
            if ( (c <= CMAX) && (!results.contains(result))) {
                results.append(result);
            }
        }
    }
}



} // GB2