/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "ProjectParsing.h"

#include <core_api/IOAdapter.h>
#include <core_api/ObjectViewModel.h>
#include <core_api/Task.h>
#include <core_api/L10n.h>
#include <core_api/AppContext.h>

#include <QtXml/qdom.h>

namespace GB2 {

static QString map2String(const QVariantMap& map) {
    QByteArray a;
    QVariant v(map);
    QDataStream s(&a, QIODevice::WriteOnly);
    s << v;
    QString res(a.toBase64());
    return res;
}

static QVariantMap string2Map(const QString& string, bool emptyMapIfError) {
    Q_UNUSED(emptyMapIfError);

    QDataStream s(QByteArray::fromBase64(string.toAscii()));
    QVariant res(QVariant::Map);
    s >> res;
    if (res.type() == QVariant::Map) {
        return res.toMap();
    }
    assert(emptyMapIfError);
    return QVariantMap();
}

static QVariant toAbsoluteRelations(const QList<GObjectRelation>& relList, const QDir& projDir)  {
    QList<GObjectRelation> absoluteRelations;
    foreach(GObjectRelation rel, relList){
        if (GUrl::getURLType(rel.ref.docUrl) == GUrl_File) { //make URL absolute
            QFileInfo ff(rel.ref.docUrl);
            if(!ff.isAbsolute()){
                QFileInfo tmpFi(projDir.path() + "/" + rel.ref.docUrl);
                rel.ref.docUrl = tmpFi.canonicalFilePath();
            }
        }
        absoluteRelations.append(rel);
    }
    return QVariant::fromValue<QList<GObjectRelation> >(absoluteRelations);
}

static QVariant toRelativeRelations(const QList<GObjectRelation>& absRelations, const QDir& projDir, 
                                    const QMap<QString, QString>& urlRemap = QMap<QString, QString>()) 
{
    QList<GObjectRelation> relativeRelations;
    foreach (const GObjectRelation& absRel, absRelations) {
        GObjectRelation relRel = absRel;
        relRel.ref.docUrl = urlRemap.value(absRel.getDocURL(), absRel.getDocURL());
        if (GUrl::getURLType(relRel.getDocURL()) == GUrl_File) { //make URL absolute

#ifdef _DEBUG //ensure internal URLs are always canonical
            QFileInfo fi(relRel.getDocURL());
            QString canonicalPath = fi.canonicalFilePath();
            assert(canonicalPath.isEmpty() || canonicalPath == relRel.getDocURL());
#endif

            QString relativeURL = projDir.relativeFilePath(relRel.getDocURL());
            relRel.ref.docUrl = relativeURL;
        }
        relativeRelations.append(relRel);
    }
    return QVariant::fromValue<QList<GObjectRelation> >(relativeRelations);
}


void ProjectFileUtils::saveProjectFile(TaskStateInfo& ts, Project* project, 
                                          const QString& projectUrl, const QMap<QString, QString>& docUrlRemap) 
{
    QDomDocument xmlDoc("GB2PROJECT");

    QDomElement projectElement = xmlDoc.createElement("gb2project");
    projectElement.setAttribute("name", project->getProjectName());
    const QString ver(CURRENT_PROJECT_VERSION);
    projectElement.setAttribute("version", ver);

    QFileInfo projectFile(projectUrl);
    QDir projectDir = projectFile.absoluteDir();

    //save documents
    foreach(Document* gbDoc, project->getDocuments()) {
        gbDoc->getDocumentFormat()->updateFormatSettings(gbDoc);

        QString docUrl = gbDoc->getURLString();
        if (docUrlRemap.contains(docUrl)) {
            docUrl = docUrlRemap.value(docUrl);
        }

        QString adapterId = gbDoc->getIOAdapterFactory()->getAdapterId();
        QDomElement docElement = xmlDoc.createElement("document");
        docElement.setAttribute("io-adapter", adapterId);

        //for local file save relative path
        if (adapterId == BaseIOAdapters::LOCAL_FILE || adapterId == BaseIOAdapters::GZIPPED_LOCAL_FILE) {
            QString relativeUrl = projectDir.relativeFilePath(docUrl);
            docElement.setAttribute("url", relativeUrl);
        } else {
            docElement.setAttribute("url", docUrl);
        }

        //store lock info for documents
        DocumentFormat* f = gbDoc->getDocumentFormat();
        QString formatId = f->getFormatId();
        docElement.setAttribute("format", formatId);
        docElement.setAttribute("readonly", gbDoc->hasUserModLock() ? 1 : 0);
        StateLock* l = gbDoc->getDocumentModLock(DocumentModLock_FORMAT_AS_INSTANCE);
        if (l!=NULL) {
            docElement.setAttribute("format-lock", 1);
        }

        //store doc-level hints
        if (!gbDoc->getGHintsMap().isEmpty()) {
            QString hintsStr = map2String(gbDoc->getGHintsMap());
            QDomText hintsNode = xmlDoc.createCDATASection(hintsStr);
            docElement.appendChild(hintsNode);
        }

        //now save unloaded objects info for all document objects 
        foreach(GObject* obj, gbDoc->getObjects()) {
            QDomElement objElement = xmlDoc.createElement("object");
            UnloadedObjectInfo info(obj);
            objElement.setAttribute("name", info.name);
            objElement.setAttribute("type", info.type);
            if (!info.hints.isEmpty()) {
                //for all object relations make path relative
                info.hints[RELATED_OBJECTS_KEY] = toRelativeRelations(obj->getObjectRelations(), projectDir, docUrlRemap);
                QString hintsStr = map2String(info.hints);
                QDomText hintsNode = xmlDoc.createCDATASection(hintsStr);
                objElement.appendChild(hintsNode);
            }
            docElement.appendChild(objElement);
        }
        projectElement.appendChild(docElement);
    }

    //save views states
    foreach(GObjectViewState* view, project->getGObjectViewStates()) {
        //save document info
        QDomElement viewElement = xmlDoc.createElement("view");
        viewElement.setAttribute("factory", view->getViewFactoryId());
        viewElement.setAttribute("viewName", view->getViewName());
        viewElement.setAttribute("stateName", view->getStateName());
        QString dataStr = map2String(view->getStateData());
        QDomText textNode = xmlDoc.createCDATASection(dataStr);
        viewElement.appendChild(textNode);

        projectElement.appendChild(viewElement);
    }

    xmlDoc.appendChild(projectElement);

    QByteArray rawData = xmlDoc.toByteArray();
    //  printf(">>%s", xmlDoc.toString().toStdString().c_str());

    //check that project directory exists
    if (!projectDir.exists()) {
        QDir root;
        root.mkpath(projectDir.absolutePath());
    }

    //save project file text to file
    QFile f(projectUrl);
    if (f.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
        qint64 s = f.write(rawData);
        f.close();
        if (s!=rawData.size()) {
            ts.setError(L10N::errorWritingFile(projectUrl));
        }
    } else {
        ts.setError(L10N::errorOpeningFileWrite(projectUrl));
    }
}



void ProjectFileUtils::loadXMLProjectModel(const QString& url, TaskStateInfo& si, QDomDocument& doc, QString& version) {
    assert(doc.isNull());

    QFile f(url);
    if (!f.open(QIODevice::ReadOnly)) {
        si.setError(  tr("error_open_file %1").arg(url) );
        return;
    }
    QByteArray  xmlData = f.readAll();
    f.close();

    bool res = doc.setContent(xmlData);

    const QString curVerStr(CURRENT_PROJECT_VERSION);

    QDomElement projectElement = doc.documentElement();
    version = projectElement.attribute("version");

    if(!version.isEmpty()){
        bool ok;
        double projVer = version.toDouble(&ok), curVer = curVerStr.toDouble();

        if (!ok) {
            si.setError(tr("Project version is not a double value"));
            doc.clear();
        }

        if (projVer > curVer){
            si.setError(tr("Project file was created by a newer version of UGENE. Please update UGENE."));
            doc.clear();
        }
    }

    if (!res || doc.doctype().name()!="GB2PROJECT") {
        si.setError(  tr("Not a valid UGENE project file %1").arg(url) );
        doc.clear();
    }
}

//////////////////////////////////////////////////////////////////////////
void ProjectParserRegistry::init(){
    parsers.qlist.append(new ProjectParser10());
}


ProjectParser* ProjectParserRegistry::getProjectParserByVersion( const QString& id ){
    foreach(ProjectParser *p,  parsers.qlist){
        if(p->getVersion() == id){
            return p;
        }
    }
    return NULL;
}

ProjectParserRegistry * ProjectParserRegistry::instance(){
    static ProjectParserRegistry ppr;
    return &ppr;
}



//////////////////////////////////////////////////////////////////////////
// Parser for v1.0 format

Project* ProjectParser10::createProjectFromXMLModel( const QString& pURL, const QDomDocument& xmlDoc, TaskStateInfo& si ){
    Q_UNUSED(si); //TODO: report about errors using si!!!

    QDomElement projectElement = xmlDoc.documentElement();
    QString name = projectElement.attribute("name");

    QList<Document*> documents;
    QList<GObjectViewState*> states;

    //read all documents
    QDomNodeList documentList = projectElement.elementsByTagName("document");
    for(int i=0; i < documentList.size(); i++) {
        QDomNode dn = documentList.item(i);
        if (!dn.isElement()) {
            continue;
        }
        QDomElement docElement = dn.toElement();
        QString ioAdapterId = docElement.attribute("io-adapter");
        QFileInfo proj(pURL);
        //QString docURL = proj.absoluteDir().path() + "/" + docElement.attribute("url");
        QString fURL(docElement.attribute("url"));
        QFileInfo fi(docElement.attribute("url"));
        QString docURL = docElement.attribute("url");
        if(ioAdapterId == BaseIOAdapters::GZIPPED_LOCAL_FILE || ioAdapterId == BaseIOAdapters::LOCAL_FILE){
            if(!fi.isAbsolute()){
                QString tmp = proj.absoluteDir().path() + "/" + docElement.attribute("url");
                QFileInfo ff(proj.absoluteDir().path() + "/" + docElement.attribute("url"));
                docURL = ff.canonicalFilePath();
            }else{
                docURL = fi.canonicalFilePath();
            }
        }
        DocumentFormatId format = docElement.attribute("format");
        bool readonly = docElement.attribute("readonly").toInt() != 0;
        bool instanceLock = docElement.attribute("format-lock").toInt() != 0;
        IOAdapterFactory* iof = AppContext::getIOAdapterRegistry()->getIOAdapterFactoryById(ioAdapterId);
        DocumentFormat* df = AppContext::getDocumentFormatRegistry()->getFormatById(format);
        assert(df!=NULL);
        QVariantMap fs = string2Map(docElement.text(), true);

        QList<UnloadedObjectInfo> unloadedObjects;
        QDomNodeList objectList = docElement.elementsByTagName("object");
        QSet<QString> objNames;
        for(int j=0; j < objectList.size(); j++) {
            QDomNode on = objectList.item(j);
            if (!on.isElement()) {
                continue;
            }
            QDomElement objElement = on.toElement();
            UnloadedObjectInfo info;
            info.name = objElement.attribute("name");
            info.type = objElement.attribute("type");
            info.hints = string2Map(objElement.text(), true);

            if (!info.hints.isEmpty()) {
                QVariant qv = info.hints[RELATED_OBJECTS_KEY];
                if (!qv.isNull()){
                    QList<GObjectRelation> relList = qv.value<QList<GObjectRelation> >();
                    info.hints[RELATED_OBJECTS_KEY] = toAbsoluteRelations(relList, proj.absoluteDir());
                }
            }

            if (info.isValid() && !objNames.contains(info.name)) {
                unloadedObjects.append(info);
                objNames.insert(info.name);
            }
        }
        Document* d = new Document(df, iof, docURL, unloadedObjects, fs, instanceLock ? tr("last_loaded_state_was_locked_by_format") : QString());
        d->setUserModLock(readonly);
        documents.append(d);
    }

    // read all saved views
    QDomNodeList viewElements = projectElement.elementsByTagName("view");
    for(int i=0;i<viewElements.size(); i++) {
        QDomNode n = viewElements.item(i);
        assert(n.isElement());
        if (!n.isElement()) {
            continue;
        }
        QDomElement viewElement = n.toElement();
        GObjectViewFactoryId id = viewElement.attribute("factory");
        QString viewName = viewElement.attribute("viewName");
        QString stateName = viewElement.attribute("stateName");
        QVariantMap map  = string2Map(viewElement.text(), false);
        GObjectViewState* state = new GObjectViewState(id, viewName, stateName, map);
        states.append(state);
    }
    Project* project = AppContext::getProjectLoader()->createProject(name, pURL, documents, states);
    return project;
}

} //namespace
