/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include <QtGui/QLabel>
#include <QtGui/QLayout>
#include <QtGui/QCheckBox>

#include <core_api/AppContext.h>
#include <core_api/CudaGpuRegistry.h>
#include <core_api/AppResources.h>

#include "CudaSupportSettingsController.h"

namespace GB2 {

CudaSupportSettingsPageController::CudaSupportSettingsPageController( const QString & _displayMsg, QObject * p /* = 0 */ ) :
AppSettingsGUIPageController( tr("CUDA Support"), CudaSupportSettingsPageId, p ), displayMsg(_displayMsg) {}

AppSettingsGUIPageState * CudaSupportSettingsPageController::getSavedState() {
    QList<CudaGpuModel *> registeredGpus = AppContext::getCudaGpuRegistry()->getRegisteredGpus();
    CudaSupportSettingsPageState * s = new CudaSupportSettingsPageState( registeredGpus.size() );
    for( int i = 0, end = s->enabledGpus.size(); i < end; ++i ) {
        s->enabledGpus[i] = registeredGpus.at(i)->isEnabled();
    }

    return s;    
}

void CudaSupportSettingsPageController::saveState( AppSettingsGUIPageState * _s ) {
    QList<CudaGpuModel *> registeredGpus = AppContext::getCudaGpuRegistry()->getRegisteredGpus();
    CudaSupportSettingsPageState * s = qobject_cast<CudaSupportSettingsPageState *>(_s);

    //saving state of enabled/disabled GPUs into registry
    for( int i = 0, end = s->enabledGpus.size(); i < end; ++i ) {
        registeredGpus[i]->setEnabled( s->enabledGpus[i] );
    }

    //increasing/decreasing maxuse of according resource
    int totalEnabled = s->enabledGpus.count(true);
    AppResource * gpuResource = AppResourcePool::instance()->getResource( RESOURCE_CUDA_GPU );
    if( gpuResource ) {
        gpuResource->maxUse = totalEnabled;
    } //else - resource was not registered, nothing to do.
}

AppSettingsGUIPageWidget * CudaSupportSettingsPageController::createWidget( AppSettingsGUIPageState* state ) {
    CudaSupportSettingsPageWidget * w = new CudaSupportSettingsPageWidget(displayMsg, this);
    w->setState(state);
    return w;
}

CudaSupportSettingsPageState::CudaSupportSettingsPageState( int num_gpus ) {
    assert( num_gpus >= 0 );
    enabledGpus.resize( num_gpus );
}

const static char * gpusDiscoveredText = "\
The following CUDA-enabled GPUs were discovered \n\
in your system. Checked units can be used by UGENE \n\
for accelerating computational algorithms.";

const static char * noGpusDiscoveredText = "No CUDA-enabled GPUs were discovered in your system.";

CudaSupportSettingsPageWidget::CudaSupportSettingsPageWidget( const QString & _msg, CudaSupportSettingsPageController * /*ctrl*/ ) :
onlyMsg(_msg){
    QVBoxLayout * vLayout = new QVBoxLayout(this);

    if( !onlyMsg.isEmpty() ) {
        //just display the centered warning message
        QLabel * msgLabel = new QLabel( onlyMsg, this );
        msgLabel->setAlignment( Qt::AlignCenter );

        vLayout->setAlignment( Qt::AlignCenter );
        vLayout->addWidget(msgLabel);
    } else {
        //everything is OK - adding info about all available GPUs
                    
        QList<CudaGpuModel *> gpus = AppContext::getCudaGpuRegistry()->getRegisteredGpus();
        const QString & actualText = gpus.empty() ? tr(noGpusDiscoveredText) : tr(gpusDiscoveredText);
        QLabel * gpusDiscoveredLabel = new QLabel( actualText, this );
        vLayout->addWidget( gpusDiscoveredLabel );

        foreach( CudaGpuModel * m, gpus ) {
            vLayout->setAlignment( Qt::AlignLeft | Qt::AlignTop );
            QHBoxLayout * hLayout = new QHBoxLayout(this);

            QString gpuText = m->getName() + " " + QString::number(m->getGlobalMemorySizeBytes() / (1024*1024)) + " Mb"; 
            QCheckBox * check = new QCheckBox( gpuText, this );

            check->setChecked(true);

            gpuEnableChecks.push_back(check);
            hLayout->addWidget(check);
            vLayout->addLayout( hLayout );
        }
    }
    setLayout(vLayout);
}

void CudaSupportSettingsPageWidget::setState( AppSettingsGUIPageState * _state ) {
    CudaSupportSettingsPageState * state = qobject_cast<CudaSupportSettingsPageState*>(_state);
    assert( state->enabledGpus.size() == gpuEnableChecks.size() );

    for( int i = 0, end = state->enabledGpus.size(); i < end; ++i ) {
        gpuEnableChecks.at(i)->setChecked( state->enabledGpus.at(i) );
    }
}

AppSettingsGUIPageState* CudaSupportSettingsPageWidget::getState( QString & /*err*/ ) const {
    CudaSupportSettingsPageState * state = new CudaSupportSettingsPageState( gpuEnableChecks.size() );

    assert( state->enabledGpus.size() == gpuEnableChecks.size() );
    for( int i = 0, end = state->enabledGpus.size(); i < end; ++i ) {
        state->enabledGpus[i] = gpuEnableChecks.at(i)->isChecked();
    }
    return state;
}

} //namespace
