/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "DotPlotClasses.h"

namespace GB2 {

DotPlotMiniMap::DotPlotMiniMap(int bigMapW, int bigMapH, float ratio) {

	this->ratio = ratio;

	Q_ASSERT(ratio>0);

	w = bigMapW/ratio;
	h = bigMapH/ratio;

	// coords of the mini map on widget
	x = bigMapW - w;
	y = bigMapH - h;
}

QRectF DotPlotMiniMap::getBoundary() const {

	return QRectF(x, y, w, h);
}

// get coordinates on the dotplot picture matching point on the minimap
QPointF DotPlotMiniMap::fromMiniMap(const QPointF &p, float zoom) const {

	float lx = p.x() - x;
	float ly = p.y() - y;

	return QPointF((lx*zoom - w/2)*ratio, (ly*zoom - h/2)*ratio);
}

// draw minimap
void DotPlotMiniMap::draw(QPainter &p, int shiftX, int shiftY, float zoom) const {

	p.save();

	p.setBrush(QBrush(QColor(200, 200, 200, 100)));

	p.translate(x, y);
	p.drawRect(0, 0, w, h);

	QRect r((-shiftX/ratio)/zoom, (-shiftY/ratio)/zoom, w/zoom, h/zoom);

	if (!r.width()) {
		r.setWidth(1);
	}
	if (!r.height()) {
		r.setHeight(1);
	}
	p.drawRect(r);

	p.restore();
}


DotPlotResultsListener::DotPlotResultsListener() {

	dotPlotList = new QList<DotPlotResults>();
	stateOk = true;
	rfTask = NULL;
}

DotPlotResultsListener::~DotPlotResultsListener() {

	delete dotPlotList;
}

void DotPlotResultsListener::setTask(Task *t) {

	rfTask = t;
	stateOk = true;
}


// add new found results to the list
void DotPlotResultsListener::onResult(const RFResult& r) {
	QMutexLocker locker(&mutex);

	if (!dotPlotList) {
		return;
	}

	if (dotPlotList->size() > maxResults) {
		Q_ASSERT(rfTask);
		rfTask->cancel();
		stateOk = false;

		return;
	}

	DotPlotResults vec(r.x, r.y, r.l);
	dotPlotList->push_back(vec);
}

void DotPlotResultsListener::onResults(const QVector<RFResult>& v) {
	QMutexLocker locker(&mutex);

	foreach (const RFResult &r, v) {

		if (!dotPlotList) {
			return;
		}

		if (dotPlotList->size() > maxResults) {
			Q_ASSERT(rfTask);
			rfTask->cancel();
			stateOk = false;

			return;
		}

		DotPlotResults vec(r.x, r.y, r.l);
		dotPlotList->push_back(vec);
	}

}

// dialogs and error messages
void DotPlotDialogs::taskRunning() {

	QMessageBox mb(QMessageBox::Critical, tr("Task is already running"), tr("Build or Load DotPlot task is already running"));
	mb.exec();
}

int DotPlotDialogs::saveDotPlot() {

	QMessageBox mb(QMessageBox::Information, tr("Save DotPlot"), tr("Save DotPlot before deletion?"), QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
	return mb.exec();
}

void DotPlotDialogs::fileOpenError(const QString &filename) {

	QMessageBox mb(QMessageBox::Critical, tr("File opening error"), tr("Error opening file %1").arg(filename));
	mb.exec();
}

void DotPlotDialogs::filesOpenError() {

	QMessageBox mb(QMessageBox::Critical, tr("Error"), tr("Error opening files"));
	mb.exec();
}

int DotPlotDialogs::loadDifferent() {

	QMessageBox mb(QMessageBox::Critical, tr("Sequences are different"), tr("Current and loading sequences are different. Continue loading DotPlot anyway?"), QMessageBox::Yes | QMessageBox::No);
	return mb.exec();
}

void DotPlotDialogs::loadWrongFormat() {

	QMessageBox mb(QMessageBox::Critical, tr("Wrong format"), tr("Wrong DotPlot file format, loading aborted"));
	mb.exec();
}

void DotPlotDialogs::wrongAlphabetTypes() {

	QMessageBox mb(QMessageBox::Critical, tr("Wrong alphabet types"), tr("Both alphabets must be default DNA nucleotide"));
	mb.exec();
}

void DotPlotDialogs::noProject() {

	QMessageBox mb(QMessageBox::Critical, tr("No project"), tr("Create or load project first"));
	mb.exec();
}

void DotPlotDialogs::tooManyResults() {

	QMessageBox mb(QMessageBox::Critical, tr("Too many results"), tr("Too many results. Try to increase min length"));
	mb.exec();
}

}
