/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "WorkflowDesignerPlugin.h"
#include "WorkflowViewController.h"
#include "WorkflowViewItems.h"
#include "WorkflowDocument.h"
#include "SceneSerializer.h"
#include "WorkflowSceneTasks.h"
#include "WorkflowPalette.h"
#include "WorkflowEditor.h"
#include "WorkflowMetaDialog.h"
#include "ChooseItemDialog.h"
#include "SchemaConfigurationDialog.h"
#include "SchemaAliasesConfigurationDialogImpl.h"
#include "WorkflowSamples.h"
#include "GlassView.h"
#include <wd_local_task/WorkflowSimpleLocalTask.h>

#include <workflow/WorkflowModel.h>
#include <workflow/WorkflowRegistry.h>
#include <workflow/IntegralBusModel.h>
#include <workflow/WorkflowEnv.h>
#include <workflow/WorkflowManager.h>
#include <workflow_support/MapDatatypeEditor.h>
#include <workflow_support/WorkflowUtils.h>
#include <workflow_support/WorkflowRunTask.h>
#include <workflow_support/WorkflowSettings.h>
#include <workflow_support/CoreDataTypes.h>
#include <workflow_support/DelegateEditors.h>
#include <workflow_support/CoreLibConstants.h>

#include <distributed_computing/RemoteMachine.h>
#include <util_gui/RemoteMachineMonitorDialogImpl.h>
#include <distributed_computing/DistributedComputingUtil.h>

#include <core_api/AppContext.h>
#include <core_api/MainWindow.h>
#include <core_api/Log.h>
#include <core_api/Settings.h>

#include <util_tasks/TaskSignalMapper.h>

#include <algorithm>
#include <functional>

#include <QtGui/QMenu>
#include <QtGui/QBoxLayout>
#include <QtGui/QSplitter>
#include <QtGui/QToolButton>

#include <QtSvg/QSvgGenerator>
#include <QtGui/QPrinter>
#include <QtGui/QPixmap>
#include <QtGui/QPainter>

#include <QtGui/QGraphicsSceneMouseEvent>
#include <QtGui/QGraphicsView>
#include <QtGui/QComboBox>
#include <QtGui/QFileDialog>
#include <QtCore/QFileInfo>
#include <QtGui/QMessageBox>
#include <QtGui/QCloseEvent>
#include <QtGui/QToolBar>
#include <QtGui/QClipboard>

#include <QtXml/qdom.h>

Q_DECLARE_METATYPE(GB2::StyleId);

/* TRANSLATOR GB2::LocalWorkflow::WorkflowView*/

namespace GB2 {

static LogCategory log(ULOG_CAT_WD);

#define LAST_DIR SETTINGS + "lastdir"
#define SPLITTER_STATE SETTINGS + "splitter"
#define EDITOR_STATE SETTINGS + "editor"
#define PALETTE_STATE SETTINGS + "palette"
#define TABS_STATE SETTINGS + "tabs"

enum {ElementsTab,SamplesTab};

#define WS 1000

static QString percentStr = WorkflowView::tr("%");

/********************************
* PercentValidator
********************************/
class PercentValidator : public QRegExpValidator {
public:
    PercentValidator(QObject* parent) : QRegExpValidator(QRegExp("[1-9][0-9]*"+percentStr), parent) {}
    void fixup(QString& input) const {
        if (!input.endsWith(percentStr)) {
            input.append(percentStr);
        }
    }
}; // PercentValidator

/********************************
* WorkflowView
********************************/
WorkflowView::WorkflowView(WorkflowGObject* go) : MWMDIWindow(tr("Workflow Designer")), go(go), currentProc(NULL), pasteCount(0) {
    scene = new WorkflowScene(this);
    scene->setSceneRect(QRectF(-WS, -WS, WS, WS));
    connect(scene, SIGNAL(processItemAdded()), SLOT(sl_procItemAdded()));
    connect(scene, SIGNAL(processDblClicked()), SLOT(sl_toggleStyle()));

    palette = new WorkflowPalette(WorkflowEnv::getProtoRegistry()/*, this*/);
    palette->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Ignored));
    connect(palette, SIGNAL(processSelected(Workflow::ActorPrototype*)), SLOT(sl_selectProcess(Workflow::ActorPrototype*)));

    infoList = new QListWidget(this);
    connect(infoList, SIGNAL(itemDoubleClicked(QListWidgetItem*)), SLOT(sl_pickInfo(QListWidgetItem*)));
    splitter = new QSplitter(this);

    tabs = new QTabWidget(this);
    tabs->insertTab(ElementsTab, palette, tr("Objects"));
    SamplesWidget* samples = new SamplesWidget();
    
    tabs->insertTab(SamplesTab, samples, tr("Samples"));
    splitter->addWidget(tabs);

    QGraphicsView* sceneView = new GlassView(scene);
    connect(samples, SIGNAL(setupGlass(GlassPane*)), sceneView, SLOT(setGlass(GlassPane*)));
    connect(samples, SIGNAL(sampleSelected(QString)), this, SLOT(sl_pasteSample(QString)));
    connect(tabs, SIGNAL(currentChanged(int)), samples, SLOT(cancelItem()));
    connect(tabs, SIGNAL(currentChanged(int)), palette, SLOT(resetSelection()));
    connect(tabs, SIGNAL(currentChanged(int)), scene, SLOT(setHint(int)));

    //sceneView->setResizeAnchor(QGraphicsView::AnchorViewCenter);
    sceneView->setAlignment(Qt::AlignCenter);
    infoSplitter = new QSplitter(Qt::Vertical, splitter);
    infoSplitter->addWidget(sceneView);
    {
        QGroupBox* w = new QGroupBox(infoSplitter);
        w->setFlat(true);
        w->setTitle(tr("Error list"));
        QVBoxLayout* vl = new QVBoxLayout(w);
        vl->setSpacing(0);
        vl->setMargin(0);
        vl->setContentsMargins(0,0,0,0);
        vl->addWidget(infoList);
        w->hide();
        infoSplitter->addWidget(w);
    }
    splitter->addWidget(infoSplitter);

    propertyEditor = new WorkflowEditor(this);
    //connect(scene, SIGNAL(selectionChanged()), propertyEditor, SLOT(clearContents()));
    //connect(scene, SIGNAL(selectionChanged()), propertyEditor, SLOT(hide()));
    connect(scene, SIGNAL(selectionChanged()), SLOT(sl_editItem()));
    splitter->addWidget(propertyEditor);

    Settings* settings = AppContext::getSettings();
    if (settings->contains(SPLITTER_STATE)) {
        splitter->restoreState(settings->getValue(SPLITTER_STATE).toByteArray());
    }
    if (settings->contains(EDITOR_STATE)) {
        propertyEditor->restoreState(settings->getValue(EDITOR_STATE));
    }
    if (settings->contains(PALETTE_STATE)) {
        palette->restoreState(settings->getValue(PALETTE_STATE));
    }
    tabs->setCurrentIndex(settings->getValue(TABS_STATE, SamplesTab).toInt());

    scene->views().at(0)->setDragMode(QGraphicsView::RubberBandDrag);
    //scene->views().at(0)->setRubberBandSelectionMode(Qt::ContainsItemShape);
    
    QHBoxLayout *layout = new QHBoxLayout;
    layout->addWidget(splitter);
    layout->setSpacing(0);
    layout->setMargin(0);
    layout->setContentsMargins(0, 0, 0, 0);
    setLayout(layout);

    createActions();
    
    if (go) {
        //FIXME
        go->setView(this);
        QMap<ActorId, ActorId> remapping;
        QString err = SceneSerializer::xml2scene(go->getXML().documentElement(), scene, remapping);
        if (!err.isEmpty()) {
            scene->sl_reset();
            log.error(err);
        } else {
            QList<Iteration> lst;
            SchemaSerializer::readIterations(lst, go->getXML().documentElement(), remapping);
            //if (!lst.isEmpty()) 
            scene->setIterations(lst);
            SchemaSerializer::readMeta(&meta, go->getXML().documentElement());
			if (go->getDocument()) {
                meta.url = go->getDocument()->getURLString();
			}
            sl_updateTitle();
            propertyEditor->resetIterations();
            scene->setModified(false);
            sl_refreshActorDocs();
        }
    } else {
        sl_newScene();
    }
    connect(scene, SIGNAL(configurationChanged()), SLOT(sl_refreshActorDocs()));
    connect(propertyEditor, SIGNAL(iterationSelected()), SLOT(sl_refreshActorDocs()));
    connect(WorkflowSettings::watcher, SIGNAL(changed()), scene, SLOT(update()));
}

WorkflowView::~WorkflowView() {
    log.trace("~WorkflowView");
}

void WorkflowView::sl_rescaleScene(const QString &scale)
{
	int percentPos = scale.indexOf(percentStr);
	double newScale = scale.left(percentPos).toDouble() / 100.0;
	QMatrix oldMatrix = scene->views().at(0)->matrix();
	scene->views().at(0)->resetMatrix();
	scene->views().at(0)->translate(oldMatrix.dx(), oldMatrix.dy());
	scene->views().at(0)->scale(newScale, newScale);
}


void WorkflowView::createActions() {
    runAction = new QAction(tr("&Run schema"), this);
    runAction->setIcon(QIcon(":workflow_designer/images/run.png"));
    runAction->setShortcut(QKeySequence("Ctrl+R"));
    connect(runAction, SIGNAL(triggered()), SLOT(sl_launch()));

    validateAction = new QAction(tr("&Validate schema"), this);
    validateAction->setIcon(QIcon(":workflow_designer/images/ok.png"));
    validateAction->setShortcut(QKeySequence("Ctrl+E"));
    connect(validateAction, SIGNAL(triggered()), SLOT(sl_validate()));

    newAction = new QAction(tr("&New schema"), this);
    newAction->setIcon(QIcon(":workflow_designer/images/filenew.png"));
	newAction->setShortcuts(QKeySequence::New);
    connect(newAction, SIGNAL(triggered()), SLOT(sl_newScene()));

    saveAction = new QAction(tr("&Save schema"), this);
    saveAction->setIcon(QIcon(":workflow_designer/images/filesave.png"));
	saveAction->setShortcuts(QKeySequence::Save);//("Ctrl+S"));
    connect(saveAction, SIGNAL(triggered()), SLOT(sl_saveScene()));

    saveAsAction = new QAction(tr("&Save schema as..."), this);
    connect(saveAsAction, SIGNAL(triggered()), SLOT(sl_saveSceneAs()));

    loadAction = new QAction(tr("&Load schema"), this);
    loadAction->setIcon(QIcon(":workflow_designer/images/fileopen.png"));
	loadAction->setShortcuts(QKeySequence::Open);//("Ctrl+L"));
    connect(loadAction, SIGNAL(triggered()), SLOT(sl_loadScene()));

    exportAction = new QAction(tr("&Export schema"), this);
    exportAction->setIcon(QIcon(":workflow_designer/images/export.png"));
    exportAction->setShortcut(QKeySequence("Ctrl+Shift+S"));
    connect(exportAction, SIGNAL(triggered()), SLOT(sl_exportScene()));

    deleteAction = new QAction(tr("Delete item"), this);
    deleteAction->setIcon(QIcon(":workflow_designer/images/delete.png"));
    deleteAction->setShortcuts(QKeySequence::Delete);
    connect(deleteAction, SIGNAL(triggered()), scene, SLOT(sl_deleteItem()));

    configureIterationsAction = new QAction(tr("Configure iterations"), this);
    configureIterationsAction->setIcon(QIcon(":workflow_designer/images/tag.png"));
    //configureIterationsAction ->setShortcut(QKeySequence::Delete);
    connect(configureIterationsAction , SIGNAL(triggered()), SLOT(sl_configureIterations()));
    
    configureAliasesAction = new QAction(tr("Configure command line aliases"), this);
    configureAliasesAction->setIcon(QIcon(":workflow_designer/images/cmdline.png"));
    connect(configureAliasesAction, SIGNAL(triggered()), SLOT(sl_configureAliases()));
    
    selectAction = new QAction(tr("Select all items"), this);
	selectAction->setShortcuts(QKeySequence::SelectAll);
    connect(selectAction, SIGNAL(triggered()), scene, SLOT(sl_selectAll()));

    bringToFrontAction = new QAction(tr("Bring to front"), this);
    bringToFrontAction->setIcon(QIcon(":workflow_designer/images/bringtofront.png"));
    connect(bringToFrontAction, SIGNAL(triggered()), scene, SLOT(sl_bringToFront()));

    sendToBackAction = new QAction(tr("Send to back"), this);
    sendToBackAction->setIcon(QIcon(":workflow_designer/images/sendtoback.png"));
    connect(sendToBackAction, SIGNAL(triggered()), scene, SLOT(sl_sendToBack()));

	copyAction = new QAction(tr("&Copy"), this);
	copyAction->setIcon(QIcon(":workflow_designer/images/editcopy.png"));
	copyAction->setShortcuts(QKeySequence::Copy);
	connect(copyAction, SIGNAL(triggered()), SLOT(sl_copyItems()));
	
	cutAction = new QAction(tr("Cu&t"), this);
	cutAction->setIcon(QIcon(":workflow_designer/images/editcut.png"));
	cutAction->setShortcuts(QKeySequence::Cut);
	connect(cutAction, SIGNAL(triggered()), SLOT(sl_cutItems()));

	pasteAction = new QAction(tr("&Paste"), this);
	pasteAction->setIcon(QIcon(":workflow_designer/images/editpaste.png"));
	pasteAction->setShortcuts(QKeySequence::Paste);
	connect(pasteAction, SIGNAL(triggered()), SLOT(sl_pasteItems()));

    sceneScaleCombo = new QComboBox(this);
	sceneScaleCombo->setEditable(true);
	sceneScaleCombo->setValidator(new PercentValidator(this));
    QStringList scales;
    scales << tr("25%") << tr("50%") << tr("75%") << tr("100%") << tr("125%") << tr("150%") << tr("200%");
    sceneScaleCombo->addItems(scales);
    sceneScaleCombo->setCurrentIndex(3);
    connect(sceneScaleCombo, SIGNAL(currentIndexChanged(const QString &)), SLOT(sl_rescaleScene(const QString &)));

    QAction* simpleStyle = new QAction(tr("Minimal"), this);
    simpleStyle->setData(QVariant(ItemStyles::SIMPLE));
    connect(simpleStyle, SIGNAL(triggered()), SLOT(sl_setStyle()));
    QAction* extStyle = new QAction(tr("Extended"), this);
    extStyle->setData(QVariant(ItemStyles::EXTENDED));
    connect(extStyle, SIGNAL(triggered()), SLOT(sl_setStyle()));
    styleActions << simpleStyle << extStyle;
    
    QAction * localHostRunMode = new QAction( tr( "Local host" ), this );
    localHostRunMode->setCheckable( true );
    localHostRunMode->setChecked( LOCAL_HOST == meta.runMode );
    connect( localHostRunMode, SIGNAL( triggered() ), SLOT( sl_setRunMode() ) );
    QAction * remoteMachineRunMode = new QAction( tr( "Remote machine" ), this );
    remoteMachineRunMode->setCheckable( true );
    remoteMachineRunMode->setChecked( REMOTE_MACHINE == meta.runMode );
    connect( remoteMachineRunMode, SIGNAL( triggered() ), SLOT( sl_setRunMode() ) );
    runModeActions << localHostRunMode << remoteMachineRunMode;
    
    unlockAction = new QAction(tr("Unlock Scene"), this);
    unlockAction->setCheckable(true);
    unlockAction->setChecked(true);
    connect(unlockAction, SIGNAL(toggled(bool)), SLOT(sl_toggleLock(bool)));
}

void WorkflowView::sl_toggleLock(bool b) {
    if (sender() != unlockAction) {
        unlockAction->setChecked(b);
        return;
    }

    if (b) {
        scene->setRunner(NULL);
    }

    deleteAction->setEnabled(b);
    selectAction->setEnabled(b);
    copyAction->setEnabled(b);
    pasteAction->setEnabled(b);
    cutAction->setEnabled(b);
    loadAction->setEnabled(b);
    //newAction->setEnabled(b);

    configureIterationsAction->setEnabled(b);
    runAction->setEnabled(b);
    validateAction->setEnabled(b);
    configureAliasesAction->setEnabled(b);
    
    propertyEditor->setEnabled(b);
    palette->setEnabled(b);

    scene->setLocked(!b);
    scene->update();
}

void WorkflowView::sl_setStyle() {
    StyleId s = qobject_cast<QAction* >(sender())->data().value<StyleId>();
    QList<QGraphicsItem*> lst = scene->selectedItems();
    if (lst.isEmpty()) {
        lst = scene->items();
    }
    foreach(QGraphicsItem* it, lst) {
        switch (it->type()) {
            case WorkflowProcessItemType:
            case WorkflowPortItemType:
            case WorkflowBusItemType:
            (static_cast<StyledItem*>(it))->setStyle(s);
        }
    }
    //update();
}

// FIXME: move to utils classes
static void removeUrlLocationParameter( Actor * actor ) {
    assert( NULL != actor );
    Attribute * attr = actor->getParameter( CoreLibConstants::URL_LOCATION_ATTR_ID );
    if( NULL != attr ) {
        Attribute * removed = actor->removeParameter( CoreLibConstants::URL_LOCATION_ATTR_ID );
        assert( attr == removed );
        delete attr;
        delete actor->getEditor()->removeDelegate( CoreLibConstants::URL_LOCATION_ATTR_ID );
    }
    URLDelegate * urlDelegate = qobject_cast<URLDelegate*>( actor->getEditor()->getDelegate( CoreLibConstants::URL_IN_ATTR_ID ) );
    if( NULL != urlDelegate ) {
        urlDelegate->sl_showEditorButton( true );
    }
}

// FIXME: move to utils classes
static void addUrlLocationParameter( Actor * actor ) {
    assert( NULL != actor );
    Attribute * urlAttr = actor->getParameter( CoreLibConstants::URL_IN_ATTR_ID );
    Attribute * urlLocationAttr = actor->getParameter( CoreLibConstants::URL_LOCATION_ATTR_ID );
    if( NULL != urlAttr && NULL == urlLocationAttr ) {
        actor->addParameter( CoreLibConstants::URL_LOCATION_ATTR_ID, 
            new Attribute( CoreLibConstants::URL_LOCATION_ATTR(), CoreDataTypes::BOOL_TYPE(), false, true ) );
        SchemaRunModeDelegate * runModeDelegate = new SchemaRunModeDelegate();
        URLDelegate * urlDelegate = qobject_cast<URLDelegate*>
            ( actor->getEditor()->getDelegate( CoreLibConstants::URL_IN_ATTR_ID ) );
        QObject::connect( runModeDelegate, SIGNAL( si_showOpenFileButton( bool ) ), urlDelegate, SLOT( sl_showEditorButton( bool ) ) );
        actor->getEditor()->addDelegate( runModeDelegate, CoreLibConstants::URL_LOCATION_ATTR_ID );
    }
}

void WorkflowView::sl_setRunMode() {
    QAction * a = qobject_cast<QAction*>( sender() );
    if( runModeActions[0] == a ) { // local host run mode
        meta.runMode = LOCAL_HOST;
    } else if( runModeActions[1] == a ) { // remote machine run mode
        meta.runMode = REMOTE_MACHINE;
    } else {
        // no sender. smbdy called it like a function
    }
    
    runModeActions[0]->setChecked( LOCAL_HOST == meta.runMode );
    runModeActions[1]->setChecked( REMOTE_MACHINE == meta.runMode );
    
    // change actors that are on pallete now
    QList<Actor*> actorsOnBoard = scene->getAllProcs();
    foreach( Actor * actor, actorsOnBoard ) {
        switch( meta.runMode ) {
        case LOCAL_HOST:
            removeUrlLocationParameter( actor );
            break;
        case REMOTE_MACHINE:
            addUrlLocationParameter( actor );
            break;
        default:
            assert( false );
        }
    }
    
    scene->sl_deselectAll();
}

void WorkflowView::sl_toggleStyle() {
    foreach(QGraphicsItem* it, scene->selectedItems()) {
        StyleId s = (static_cast<StyledItem*>(it))->getStyle();
        if (s == ItemStyles::SIMPLE) {
            s = ItemStyles::EXTENDED;
        } else {
            s = ItemStyles::SIMPLE;
        }
        (static_cast<StyledItem*>(it))->setStyle(s);
    }
    //update();
}

void WorkflowView::sl_refreshActorDocs() {
    foreach(QGraphicsItem* it, scene->items()) {
        if (it->type() == WorkflowProcessItemType) {
            Actor* a = qgraphicsitem_cast<WorkflowProcessItem*>(it)->getProcess();
            a->getDescription()->update(propertyEditor->getCurrentIteration().getParameters(a->getId()));
        }
    }
}


void WorkflowView::setupMDIToolbar(QToolBar* tb) {
    tb->addAction(newAction);
    tb->addAction(loadAction);
    tb->addAction(saveAction);
    tb->addSeparator();
    tb->addAction(validateAction);
    tb->addAction(runAction);
    tb->addAction(configureIterationsAction);
    tb->addAction(configureAliasesAction);
    tb->addSeparator();
	tb->addAction(copyAction);
	tb->addAction(pasteAction);
	tb->addAction(cutAction);
    tb->addAction(deleteAction);
    tb->addSeparator();
    tb->addAction(bringToFrontAction);
    tb->addAction(sendToBackAction);
    tb->addSeparator();
    tb->addWidget(sceneScaleCombo);

    QToolButton* tt = new QToolButton(tb);
    QMenu* ttMenu = new QMenu(tr("Item style"), this);
    foreach(QAction* a, styleActions) {
        ttMenu->addAction(a);
    }
    tt->setDefaultAction(ttMenu->menuAction());
    tt->setPopupMode(QToolButton::InstantPopup);
    tb->addWidget(tt);
    
    QToolButton * runModeToolButton = new QToolButton(tb);
    QMenu * runModeMenu = new QMenu( tr( "Run mode" ), this );
    foreach( QAction * a, runModeActions ) {
        runModeMenu->addAction( a );
    }
    runModeToolButton->setDefaultAction( runModeMenu->menuAction() );
    runModeToolButton->setPopupMode( QToolButton::InstantPopup );
    tb->addWidget( runModeToolButton );
}

void WorkflowView::setupViewMenu(QMenu* m) {
    m->addMenu(palette->createMenu(tr("Add item")));
	m->addAction(copyAction);
	m->addAction(pasteAction);
	m->addAction(cutAction);
    m->addAction(deleteAction);
    m->addAction(selectAction);
    m->addSeparator();
    m->addAction(bringToFrontAction);
    m->addAction(sendToBackAction);
    m->addSeparator();
    m->addAction(newAction);
    m->addAction(loadAction);
    m->addAction(saveAction);
    m->addAction(saveAsAction);
    m->addAction(exportAction);
    m->addSeparator();
    m->addAction(validateAction);
    m->addAction(configureIterationsAction);
    m->addAction(configureAliasesAction);
    m->addAction(runAction);
    m->addSeparator();

    QMenu* ttMenu = new QMenu(tr("Item style"));
    foreach(QAction* a, styleActions) {
        ttMenu->addAction(a);
    }
    m->addMenu(ttMenu);
    
    QMenu * runModeMenu = new QMenu( tr( "Run mode" ) );
    foreach( QAction * a, runModeActions ) {
        runModeMenu->addAction( a );
    }
    m->addMenu( runModeMenu );
    
    if (!unlockAction->isChecked()) {
        m->addSeparator();
        m->addAction(unlockAction);
    }
}

void WorkflowView::setupContextMenu(QMenu* m) {
    if (!unlockAction->isChecked()) {
        m->addAction(unlockAction);
        return;
    }
    if (!QApplication::clipboard()->text().isEmpty()) {
        m->addAction(pasteAction);
    }
    QList<QGraphicsItem*> sel = scene->selectedItems();
    if (!sel.isEmpty()) {
		m->addAction(copyAction);
		m->addAction(cutAction);
        m->addAction(deleteAction);
        m->addSeparator();
        m->addAction(bringToFrontAction);
        m->addAction(sendToBackAction);
        m->addSeparator();
        if (sel.size() == 1 && sel.first()->type() == WorkflowProcessItemType) {
            QMenu* itMenu = new QMenu(tr("Item properties"));
            WorkflowProcessItem* wit = qgraphicsitem_cast<WorkflowProcessItem*>(sel.first());
            foreach(QAction* a, wit->getContextMenuActions()) {
                itMenu->addAction(a);
            }
            m->addMenu(itMenu);
        }
        
        QMenu* ttMenu = new QMenu(tr("Item style"));
        foreach(QAction* a, styleActions) {
            ttMenu->addAction(a);
        }
        m->addMenu(ttMenu);
        m->addSeparator();
        
        QMenu * runModeMenu = new QMenu( tr( "Run mode" ) );
        foreach( QAction * a, runModeActions ) {
            runModeMenu->addAction( a );
        }
        m->addMenu( runModeMenu );
        m->addSeparator();
    }
    m->addAction(selectAction);
    m->addMenu(palette->createMenu(tr("Add item")));
}

void WorkflowView::sl_pickInfo(QListWidgetItem* info) {
    ActorId id = info->data(ACTOR_REF).value<ActorId>();
    foreach(QGraphicsItem* it, scene->items()) {
        if (it->type() == WorkflowProcessItemType)
        {
            WorkflowProcessItem* proc = static_cast<WorkflowProcessItem*>(it);
            if (proc->getProcess()->getId() != id) {
                continue;
            } 
            scene->clearSelection();
            QString pid = info->data(PORT_REF).toString();
            WorkflowPortItem* port = proc->getPort(pid);
            if (port) {
                port->setSelected(true);
            } else {
                proc->setSelected(true);
                int itid = info->data(ITERATION_REF).toInt();
                propertyEditor->selectIteration(itid);
            }
            return;
        }
    }
}

bool WorkflowView::sl_validate(bool notify) {
    propertyEditor->commit();
    infoList->clear();
    QList<QListWidgetItem*> lst;
    bool good = DesignerUtils::validate(scene->getSchema(), &lst);

    if (lst.count() != 0) {
        foreach(QListWidgetItem* wi, lst) {
            infoList->addItem(wi);
        }
        infoList->parentWidget()->show();
        QList<int> s = infoSplitter->sizes();
        if (s.last() == 0) {
            s.last() = qMin(infoList->sizeHint().height(), 300);
            infoSplitter->setSizes(s);
        }
    } else {
        infoList->parentWidget()->hide();
    }
    if (!good) {

        QMessageBox::warning(this, tr("Schema cannot be executed"), 
            tr("Please fix issues listed in the error list (located under schema)."));
    } else {
        if (notify) {
            QMessageBox::information(this, tr("Schema is valid"), 
                tr("Schema is valid.\nWell done!"));
        }
    }
    return good;
}

void WorkflowView::localHostLaunch() {
    if (!sl_validate(false)) {
        return;
    }
    Schema sh = scene->getSchema();
    if (sh.domain.isEmpty()) {
        //|TODO user choice
        sh.domain = WorkflowEnv::getDomainRegistry()->getAllIds().value(0);
    }
    
    WorkflowRunTask* t = new WorkflowRunTask(sh, scene->getIterations());
    t->setReportingEnabled(true);
    if (WorkflowSettings::monitorRun()) {
        unlockAction->setChecked(false);
        scene->setRunner(t);
        connect(t, SIGNAL(si_ticked()), scene, SLOT(update()));
        connect(new TaskSignalMapper(t), SIGNAL(si_taskFinished(Task*)), SLOT(sl_toggleLock()));
    }
    AppContext::getTaskScheduler()->registerTopLevelTask(t);
}

void WorkflowView::remoteLaunch() {
    if( !sl_validate(false) ) {
        return;
    }
    Schema sh = scene->getSchema();
    if( sh.domain.isEmpty() ) {
        sh.domain = WorkflowEnv::getDomainRegistry()->getAllIds().value( 0 );
    }
    
    int ret = 0;
    RemoteMachineSettings * settings = NULL;
    do {
        RemoteMachineMonitor * rmm = AppContext::getRemoteMachineMonitor();
        assert( NULL != rmm );
        RemoteMachineMonitorDialogImpl dlg( QApplication::activeWindow(), rmm->getRemoteMachineMonitorItems(), 
            WorkflowSimpleLocalTaskFactory::ID );
        ret = dlg.exec();
        if( QDialog::Rejected == ret ) {
            return;
        }
        assert( QDialog::Accepted == ret );
        QList< RemoteMachineMonitorDialogItem > dlgModel = dlg.getModel();
        DistributedComputingUtil::applyChangesForRemoteMachineMonitor( rmm, dlgModel );
        QList< RemoteMachineSettings* > selectedMachines = rmm->getSelectedMachines();
        int howManyMachines = selectedMachines.size();
        if( 0 == howManyMachines ) {
            QMessageBox::critical( this, tr( "Selecting machines error!" ), tr( "You didn't select a machine to run remote task!" ) );
            continue;
        } else if( 1 != howManyMachines ) {
            QMessageBox::critical( this, tr( "Selecting machines error!" ), 
                tr( "Distributed run on many machines is not supported yet. Select 1 machine" ) );
            continue;
        }
        assert( 1 == howManyMachines );
        settings = selectedMachines.first();
        break;
    } while( QDialog::Accepted == ret );
    
    AppContext::getTaskScheduler()->registerTopLevelTask( new WorkflowRemoteRunTask( settings, sh, scene->getIterations() ) );
}

void WorkflowView::sl_launch() {
    switch( meta.runMode ) {
    case LOCAL_HOST:
        localHostLaunch();
        break;
    case REMOTE_MACHINE:
        remoteLaunch();
        break;
    default:
        assert( false );
    }
}

void WorkflowView::sl_configureIterations() {
    propertyEditor->commit();
    SchemaConfigurationDialog d(scene->getSchema(), scene->getIterations(), this);
    int ret = d.exec();
    if (d.hasModifications()) {
        scene->setIterations(d.getIterations());
        propertyEditor->resetIterations();
    }
    if (QDialog::Accepted == ret) {
        sl_launch();
    }
}

static Actor* findActorById( const Schema & sc, const ActorId & id ) {
    Actor * ret = NULL;
    foreach( Actor * a, sc.procs ) {
        assert( a != NULL );
        if( id == a->getId() ) {
            return a;
        }
    }
    return ret;
}

void WorkflowView::sl_configureAliases() {
    SchemaAliasesConfigurationDialogImpl dlg( scene->getSchema(), this );
    int ret = QDialog::Accepted;
    do {
        ret = dlg.exec();
        if( ret == QDialog::Accepted ) {
            if( !dlg.validateModel() ) {
                QMessageBox::critical( this, tr("Bad input!"), tr("Aliases for schema parameters should be different!") );
                continue;
            }
            SchemaAliasesCfgDlgModel model = dlg.getModel();
            SchemaAliasesCfgDlgModel::const_iterator it = model.constBegin();
            while( it != model.constEnd() ) {
                QMap<Descriptor, QString>::const_iterator aliasIt = it.value().constBegin();
                while( aliasIt != it.value().constEnd() ) {
                    assert( !aliasIt.value().isEmpty() );
                    Actor * actor = findActorById( scene->getSchema(), it.key() );
                    assert( actor != NULL );
                    actor->getParamAliases().insert( aliasIt.key().getId(), aliasIt.value() );
                    ++aliasIt;
                }
                ++it;
            }
            break;
        } else if( ret == QDialog::Rejected ) {
            break;
        } else {
            assert(false);
        }
    } while( ret == QDialog::Accepted );
}

void WorkflowView::sl_selectProcess(Workflow::ActorPrototype* p) {
    currentProc = p;
    propertyEditor->reset();
    if (!p) {
        //log.debug("deselected");
        scene->views().at(0)->unsetCursor();
    } else {
        propertyEditor->setDescriptor(p, tr("Drag the palette item to scene or just click on the scene to create a new object."));
        //log.debug(p->getDisplayName() + " selected");
        scene->views().at(0)->setCursor(Qt::CrossCursor);
    }
}

void WorkflowView::sl_copyItems() {
	QList<QGraphicsItem*> items = scene->selectedItems();
	if (items.isEmpty()) return;
	QDomDocument xml(SchemaSerializer::WORKFLOW_DOC);
	QDomElement root = xml.createElement("root");
	xml.appendChild(root);
	SceneSerializer::saveItems(items, root);
    QList<Iteration> lst;
    foreach(const Iteration& it, scene->getIterations()) {
        Iteration copy(it.name);
        copy.id = it.id;
        foreach(QGraphicsItem* item, items) {
            WorkflowProcessItem* proc = qgraphicsitem_cast<WorkflowProcessItem*>(item);
            if (proc && it.cfg.contains(proc->getProcess()->getId())) {
                copy.cfg.insert(proc->getProcess()->getId(), it.cfg.value(proc->getProcess()->getId()));
            }
        }
        if (!copy.cfg.isEmpty()) {
            lst.append(copy);
        }
    }
    if (!lst.isEmpty()) {
        SchemaSerializer::saveIterations(lst, root);
    }
    lastPaste = xml.toString();
    QApplication::clipboard()->setText(lastPaste);
    pasteCount = 0;
}

void WorkflowView::sl_cutItems() {
	sl_copyItems();
	scene->sl_deleteItem();
}

void WorkflowView::sl_pasteSample(const QString& s) {
    tabs->setCurrentIndex(ElementsTab);
    QDomDocument xml;
    xml.setContent(s);
   if (scene->items().isEmpty()) {
        sl_pasteItems(s);
        SchemaSerializer::readMeta(&meta, xml.documentElement());
        sl_setRunMode();
        sl_updateTitle();
    } else {
        saveState();
        WorkflowGObject* smpl = new WorkflowGObject("Sample", xml);
        WorkflowView* view = new WorkflowView(smpl);
        view->sl_setRunMode();
        AppContext::getMainWindow()->getMDIManager()->addMDIWindow(view);
        AppContext::getMainWindow()->getMDIManager()->activateWindow(view);
    }
}

void WorkflowView::sl_pasteItems(const QString& s) {
    QString tmp = s.isNull() ? QApplication::clipboard()->text() : s;
    if (tmp == lastPaste) {
        ++pasteCount;
    } else {
        pasteCount = 0;
        lastPaste = tmp;
    }
    QDomDocument xml;
    if (!xml.setContent(lastPaste) || xml.doctype().name() != SchemaSerializer::WORKFLOW_DOC) {
        return;
    }
    scene->clearSelection();
    QMap<ActorId, ActorId> remapping;
    QString msg = SceneSerializer::xml2scene(xml.documentElement(), scene, remapping, true, true);
    if (!msg.isEmpty()) {
        log.trace("Paste issues: " + msg);
    }
    // merge iteration data
    QList<Iteration> iterations;
    SchemaSerializer::readIterations(iterations, xml.documentElement(), remapping);
    if (!iterations.isEmpty()) {
        QList<Iteration> current = scene->getIterations();
        if (iterations.size() == 1 || current.size() == 1) {
            // one-to-many mapping
            Iteration& it = iterations.size() == 1 ? iterations[0] : current[0];
            QList<Iteration>& target = iterations.size() == 1 ? current : iterations;
            for (int i = 0; i < target.size(); i++)
            {
                target[i].cfg.unite(it.cfg); //unite is safe as no actor ids can coincide
            }
            scene->setIterations(target);
        } else {
            // many-to-many
            foreach(const Iteration& pasted, iterations) {
                bool gotIt = false;
                for (int i = 0; i < current.size(); i++)
                {
                    Iteration& it = current[i];
                    if (it.name == pasted.name && it.id == pasted.id) {
                        it.cfg.unite(pasted.cfg);
                        gotIt = true;
                        break;
                    } 
                }
                if (!gotIt) {
                    current.append(pasted);
                }
            }
            scene->setIterations(current);
        }
        propertyEditor->resetIterations();
    }

	QGraphicsItemGroup* g = scene->createItemGroup(scene->selectedItems());
    int shift = GRID_STEP*(pasteCount);
	g->moveBy(shift, shift);
	scene->destroyItemGroup(g);
}

void WorkflowView::sl_procItemAdded() {
    if (!currentProc) return;
    log.trace(currentProc->getDisplayName() + " added");
    palette->resetSelection();
    currentProc = NULL;
    assert(scene->views().size() == 1);
    scene->views().at(0)->unsetCursor();
}

void WorkflowView::sl_showEditor() {
    propertyEditor->show();
    QList<int> s = splitter->sizes();
    if (s.last() == 0) {
        s.last() = propertyEditor->sizeHint().width();
        splitter->setSizes(s);
    }
}

void WorkflowView::sl_editItem() {
    QList<QGraphicsItem*> list = scene->selectedItems();
    if (list.size() == 1) {
        QGraphicsItem* it = list.at(0);
        if (it->type() == WorkflowProcessItemType) {
            propertyEditor->editActor(qgraphicsitem_cast<WorkflowProcessItem*>(it)->getProcess());
            return;
        }
        Port* p = NULL;
        if (it->type() == WorkflowBusItemType) {
            p = qgraphicsitem_cast<WBusItem*>(it)->getInPort()->getPort();
        } else if (it->type() == WorkflowPortItemType) {
            p = qgraphicsitem_cast<WorkflowPortItem*>(it)->getPort();
        }
        if (p) {
            if (qobject_cast<BusPort*>(p)) 
            {
                BusPortEditor* ed = new BusPortEditor(qobject_cast<BusPort*>(p));
                ed->setParent(p);
                p->setEditor(ed);
            }
        }
        propertyEditor->editPort(p);
    } else {
        propertyEditor->reset();
    }
}

void WorkflowView::sl_exportScene() {
    propertyEditor->commit();

    // TODO more export options and features
    QString dir = AppContext::getSettings()->getValue(LAST_DIR, QString("")).toString();
    QString filter = tr("Raster image (*.png *.bmp *.jpg *.jpeg *.ppm *.xbm *.xpm)");
    filter += "\n" + tr("Vector image (*.svg)");
    filter += "\n"+ tr("Portable document (*.pdf *.ps)");
    QString selectedFilter = "";//AppContext::getSettings()->getValue(LAST_FILE_FILTER, QString("")).toString();
    QString url = QFileDialog::getSaveFileName(0, tr("Export workflow schema to image"), dir, filter,&selectedFilter);
    if (!url.isEmpty()) {
        AppContext::getSettings()->setValue(LAST_DIR, QFileInfo(url).absoluteDir().absolutePath());
        log.details(tr("Saving scene image to file: %1").arg(url));
        bool result = false;
        QRectF bounds = scene->itemsBoundingRect();
        if (url.endsWith(".svg", Qt::CaseInsensitive)) {
            QSvgGenerator svg;
            svg.setFileName(url);
            svg.setSize(bounds.size().toSize());
            QPainter painter(&svg);
            painter.setRenderHint(QPainter::Antialiasing);
            scene->render(&painter, QRectF(), bounds);
            result = painter.end();
        } else if (url.endsWith(".pdf", Qt::CaseInsensitive) || url.endsWith(".ps", Qt::CaseInsensitive)) {
            QPrinter printer;
            printer.setOutputFileName(url);
            QPainter painter(&printer);
            painter.setRenderHint(QPainter::Antialiasing);
            scene->render(&painter, QRectF(), bounds);
            result = painter.end();
        } else {
            QPixmap pixmap(bounds.size().toSize());
            QPainter painter(&pixmap);
            painter.setRenderHint(QPainter::Antialiasing);
            scene->render(&painter, QRectF(), bounds);
            result = painter.end() & pixmap.save(url);
        }
        if (!result) {
            log.error(WorkflowView::tr("failed to save image to %1").arg(url));
        }
    }
}

void WorkflowView::sl_saveScene() {
    if (meta.url.isEmpty()) {
        WorkflowMetaDialog md(meta);
        int rc = md.exec();
        if (rc != QDialog::Accepted) {
            return;
        }
        meta = md.meta;
        sl_updateTitle();
    }
    propertyEditor->commit();
    Task* t = new SaveWorkflowSceneTask(scene, meta); 
    AppContext::getTaskScheduler()->registerTopLevelTask(t);
}

void WorkflowView::sl_saveSceneAs() {
    WorkflowMetaDialog md(meta);
    int rc = md.exec();
    if (rc != QDialog::Accepted) {
        return;
    }
    propertyEditor->commit();
    meta = md.meta;
    Task* t = new SaveWorkflowSceneTask(scene, meta);
    AppContext::getTaskScheduler()->registerTopLevelTask(t);
    sl_updateTitle();
}

void WorkflowView::sl_loadScene() {
    if (!confirmModified()) {
        return;
    }

    QString dir = AppContext::getSettings()->getValue(LAST_DIR, QString("")).toString();
    QString filter = DesignerUtils::getSchemaFileFilter(true);
    QString url = QFileDialog::getOpenFileName(0, tr("Open workflow schema file"), dir, filter);
    if (!url.isEmpty()) {
        AppContext::getSettings()->setValue(LAST_DIR, QFileInfo(url).absoluteDir().absolutePath());
        Task* t = new LoadWorkflowSceneTask(scene, &meta, url); //FIXME unsynchronized meta usage
        TaskSignalMapper* m = new TaskSignalMapper(t);
        connect(m, SIGNAL(si_taskFinished(Task*)), SLOT(sl_updateTitle()));
        connect(m, SIGNAL(si_taskFinished(Task*)), scene, SLOT(centerView()));
        connect(m, SIGNAL(si_taskFinished(Task*)), propertyEditor, SLOT(resetIterations()));
        AppContext::getTaskScheduler()->registerTopLevelTask(t);
    }
}

void WorkflowView::sl_newScene() {
    if (!confirmModified()) {
        return;
    }
    infoList->parentWidget()->hide();
    scene->sl_reset();
    meta.reset();
    meta.name = tr("New schema");
    sl_updateTitle();
    propertyEditor->resetIterations();
    scene->setModified(false);
}

void WorkflowView::sl_updateTitle() {
    setWindowName(tr("Workflow Designer - %1").arg(meta.name));    
}

void WorkflowView::saveState() {
    AppContext::getSettings()->setValue(SPLITTER_STATE, splitter->saveState());
    AppContext::getSettings()->setValue(EDITOR_STATE, propertyEditor->saveState());
    AppContext::getSettings()->setValue(PALETTE_STATE, palette->saveState());
    AppContext::getSettings()->setValue(TABS_STATE, tabs->currentIndex());
}

bool WorkflowView::onCloseEvent() {
    saveState();
    if (!confirmModified()) {
        return false;
    }
    if (go) {
        go->setView(NULL);
    }
    return true;
}

bool WorkflowView::confirmModified() {
    propertyEditor->commit();
    if (scene->isModified() && !scene->items().isEmpty()) {
        AppContext::getMainWindow()->getMDIManager()->activateWindow(this);
        int ret = QMessageBox::question(this, tr("Workflow Designer"),
            tr("The schema has been modified.\n"
            "Do you want to save changes?"),
            QMessageBox::Save | QMessageBox::Discard
            | QMessageBox::Cancel,
            QMessageBox::Save);
        if (QMessageBox::Cancel == ret) {
            return false;
        } else if (QMessageBox::Discard == ret) {
            //scene->setModified(false);
        } else {
            sl_saveScene();
        }
    }
    return true;
}

/********************************
 * WorkflowScene
 ********************************/
static bool canDrop(const QMimeData* m, QList<ActorPrototype*>& lst) {
    if (m->hasFormat(WorkflowPalette::MIME_TYPE)) {
        QString id(m->data(WorkflowPalette::MIME_TYPE));
        ActorPrototype* proto = WorkflowEnv::getProtoRegistry()->getProto(id);
        if (proto) {
            lst << proto;
        }
    } else {
        foreach(QList<ActorPrototype*> l, WorkflowEnv::getProtoRegistry()->getProtos().values()) {
            foreach(ActorPrototype* proto, l) {
                if (proto->isAcceptableDrop(m)) {
                    lst << proto;
                }
            }
        }
    }
    //foreach(ActorPrototype* a, lst) {log.debug("drop acceptable: " + a->getId());}
    return !lst.isEmpty();
}

WorkflowScene::WorkflowScene(WorkflowView *parent) 
: QGraphicsScene(parent), controller(parent), modified(false), locked(false), runner(NULL), hint(0), nextActorNumber(0) {
}

WorkflowScene::~WorkflowScene() {
    sl_reset();
}

Schema WorkflowScene::getSchema() const {
    Schema schema;
    foreach(QGraphicsItem* it, items()) {
        if (it->type() == WorkflowProcessItemType) 
        {
            schema.procs << (static_cast<WorkflowProcessItem*>(it))->getProcess();
        } 
        else if (it->type() == WorkflowBusItemType) 
        {
            schema.flows << (static_cast<WBusItem*>(it))->getBus();
        }
    }
    schema.iterations = iterations;

    return schema;
}

void WorkflowScene::sl_deleteItem() {
    assert(!locked);
    QList<ActorId> ids;
    QList<QGraphicsItem*> items;
    foreach(QGraphicsItem* it, selectedItems()) {
        switch (it->type()) {
            case WorkflowProcessItemType:
                ids << qgraphicsitem_cast<WorkflowProcessItem*>(it)->getProcess()->getId();
                items << it;
        }
    }
    modified |= !items.isEmpty();
    foreach(QGraphicsItem* it, items) {
        removeItem(it);
        delete it;
    }
    foreach(QGraphicsItem* it, selectedItems()) {
        switch (it->type()) {
            case WorkflowBusItemType:
                removeItem(it);
                delete it;
                modified = true;
                break;
        }
    }
    bool cfgModified = false;
    for (int i = 0; i<iterations.size(); i++) {
        Iteration& it = iterations[i];
        foreach(const ActorId& id, ids) {
            if (it.cfg.contains(id)) {
                it.cfg.remove(id);
                cfgModified = true;
            }
        }
    }
    if (cfgModified) {
        controller->propertyEditor->resetIterations();
        emit configurationChanged();
    }
    update();
}

QList<Actor*> WorkflowScene::getSelectedProcItems() const {
    QList<Actor*> list;
    foreach (QGraphicsItem *item, selectedItems()) {
        if (item->type() == WorkflowProcessItemType) {
            list << static_cast<WorkflowProcessItem*>(item)->getProcess();
        }
    }
    return list;
}

QList<Actor*> WorkflowScene::getAllProcs() const {
    QList<Actor*> res;
    foreach( QGraphicsItem * item, items() ) {
        assert( NULL != item );
        if (item->type() == WorkflowProcessItemType) {
            res << static_cast<WorkflowProcessItem*>(item)->getProcess();
        }
    }
    return res;
}

void WorkflowScene::contextMenuEvent(QGraphicsSceneContextMenuEvent * e) {
    QMenu menu;
    controller->setupContextMenu(&menu);
    e->accept();
    menu.exec(e->screenPos());
}

void WorkflowScene::mouseDoubleClickEvent(QGraphicsSceneMouseEvent * mouseEvent) {
    if (!mouseEvent->isAccepted() && (mouseEvent->button() == Qt::LeftButton) && !selectedItems().isEmpty()) {
        emit processDblClicked();
    }
    QGraphicsScene::mousePressEvent(mouseEvent);
}

void WorkflowScene::dragEnterEvent(QGraphicsSceneDragDropEvent * event) {
    QList<ActorPrototype*> lst;
    if (!locked && canDrop(event->mimeData(), lst)) {
        event->acceptProposedAction();
    } else {
        QGraphicsScene::dragEnterEvent(event);
    }
}

void WorkflowScene::dragMoveEvent(QGraphicsSceneDragDropEvent * event) {
    QList<ActorPrototype*> lst;
    if (!locked && canDrop(event->mimeData(), lst)) {
        event->acceptProposedAction();
    } else {
        QGraphicsScene::dragMoveEvent(event);
    }
}

void WorkflowScene::dropEvent(QGraphicsSceneDragDropEvent * event) {
    QList<ActorPrototype*> lst;
    if (!locked && canDrop(event->mimeData(), lst))
    {
        QList<QGraphicsItem *> targets = items(event->scenePos());
        bool done = false;
        foreach(QGraphicsItem* it, targets) {
            WorkflowProcessItem* target = qgraphicsitem_cast<WorkflowProcessItem*>(it);
            if (target && lst.contains(target->getProcess()->getProto())) {
                clearSelection();
                QVariantMap params;
                Actor* a = target->getProcess();
                a->getProto()->isAcceptableDrop(event->mimeData(), &params);
                QMapIterator<QString, QVariant> cfg(params);
                while (cfg.hasNext())
                {
                    cfg.next();
                    a->setParameter(cfg.key(),cfg.value());
                }
                target->setSelected(true);
                done = true;
                break;
            }
        }
        if (!done) {
            ActorPrototype* proto = lst.size() > 1 ? ChooseItemDialog(controller).select(lst) : lst.first();
            if (proto) {
                QVariantMap params;
                proto->isAcceptableDrop(event->mimeData(), &params);
                addProcess( createActor( proto, params), event->scenePos());
                event->setDropAction(Qt::CopyAction);
            }
        }
    } 
    QGraphicsScene::dropEvent(event);
}


void WorkflowScene::addProcess(Actor* proc, const QPointF& pos) {
    WorkflowProcessItem* it = new WorkflowProcessItem(proc);
    it->setPos(pos);
    addItem(it);
    modified = true;

    emit processItemAdded();
    
    //if(items().isEmpty()) {
        // erase drop hint
    update();
    //}
}

Actor * WorkflowScene::createActor( ActorPrototype * proto, const QVariantMap & params ) {
    assert( NULL != proto );
    Actor * actor = proto->createInstance( params );
    assert( NULL != actor );
    
    nextActorNumber++;
    actor->setLabel(proto->getDisplayName() + QString(" %1").arg(nextActorNumber));
    actor->setNumberInSchema(nextActorNumber);
    
    if( REMOTE_MACHINE == controller->getMeta().runMode ) {
        addUrlLocationParameter( actor );
    }
    return actor;
}

void WorkflowScene::mousePressEvent(QGraphicsSceneMouseEvent *mouseEvent) {
    if (!locked && !mouseEvent->isAccepted() && controller->selectedProto() && (mouseEvent->button() == Qt::LeftButton)) {
        addProcess( createActor( controller->selectedProto() ), mouseEvent->scenePos());
    }
    QGraphicsScene::mousePressEvent(mouseEvent);
}

void WorkflowScene::mouseMoveEvent(QGraphicsSceneMouseEvent *mouseEvent) {
    QGraphicsScene::mouseMoveEvent(mouseEvent);
}
void WorkflowScene::mouseReleaseEvent(QGraphicsSceneMouseEvent *mouseEvent) {
    QGraphicsScene::mouseReleaseEvent(mouseEvent);
}

void WorkflowScene::sl_bringToFront()
{
    if (selectedItems().isEmpty())
        return;

    foreach(QGraphicsItem *selectedItem, selectedItems()) {
        qreal zValue = selectedItem->zValue();
        foreach (QGraphicsItem *item, selectedItem->collidingItems()) {
            if (item->zValue() >= zValue /*&& item->type() == WorkflowProcessItemType*/)
                zValue = item->zValue() + 1;
        }
        if (selectedItem->zValue() != zValue) {
            selectedItem->setZValue(zValue);
            modified = true;
        }
    }
}

void WorkflowScene::sl_sendToBack()
{
    if (selectedItems().isEmpty())
        return;
    foreach(QGraphicsItem *selectedItem, selectedItems()) {
        qreal zValue = selectedItem->zValue();
        foreach (QGraphicsItem *item, selectedItem->collidingItems()) {
            if (item->zValue() <= zValue /*&& item->type() == WorkflowProcessItemType*/)
                zValue = item->zValue() - 1;
        }
        if (selectedItem->zValue() != zValue) {
            selectedItem->setZValue(zValue);
            modified = true;
        }
    }
}

void WorkflowScene::sl_selectAll() {
    foreach(QGraphicsItem* it, items()) {
        //if (it->type() == WorkflowProcessItemType) 
        {
            it->setSelected(true);
        }
    }
}

void WorkflowScene::sl_deselectAll() {
    foreach(QGraphicsItem* it, items()) {
        it->setSelected( false );
    }
}

void WorkflowScene::sl_reset() {
    QList<QGraphicsItem*> list;
    foreach(QGraphicsItem* it, items()) {
        if (it->type() == WorkflowProcessItemType) {
            list << it;
        }
    }
    modified = false;
    foreach(QGraphicsItem* it, list) {
        removeItem(it);
        delete it;
    }
    iterations.clear();
}

void WorkflowScene::drawBackground(QPainter * painter, const QRectF & rect)
{
    if (WorkflowSettings::showGrid()) {
        int step = GRID_STEP;
        painter->setPen(QPen(QColor(200, 200, 255, 125)));
        // draw horizontal grid
        qreal start = round(rect.top(), step);
        if (start > rect.top()) {
            start -= step;
        }
        for (qreal y = start - step; y < rect.bottom(); ) {
            y += step;
            painter->drawLine(rect.left(), y, rect.right(), y);
        }
        // now draw vertical grid
        start = round(rect.left(), step);
        if (start > rect.left()) {
            start -= step;
        }
        for (qreal x = start - step; x < rect.right(); ) {
            x += step;
            painter->drawLine(x, rect.top(), x, rect.bottom());
        }
    }

    if (items().size() == 0) {
        // draw a hint on empty scene
        painter->setPen(Qt::darkGray);
        QFont f = painter->font();
        if (hint == SamplesTab) {
        } else {
            QTransform trans = painter->combinedTransform();
            f.setFamily("Courier New");
            f.setPointSizeF(f.pointSizeF()* 2./trans.m11());
            painter->setFont(f);
            QRectF res;
            painter->drawText(sceneRect(), Qt::AlignCenter, tr("Drop here an object from palette"), &res);
            QPixmap pix(":workflow_designer/images/leftarrow.png");
            QPointF pos(res.left(), res.center().y());
            pos.rx() -= pix.width() + GRID_STEP;
            pos.ry() -= pix.height()/2;
            painter->drawPixmap(pos, pix);
        }
    }
}

void WorkflowScene::setIterations( const QList<Iteration>& lst )
{
    assert(!locked);
    iterations = lst; 
    modified = true; 
    emit configurationChanged();
}

void WorkflowScene::centerView() {
    QRectF childRect;
    foreach (QGraphicsItem *child, items()) {
        QPointF childPos = child->pos();
        QTransform matrix = child->transform() * QTransform().translate(childPos.x(), childPos.y());
        childRect |= matrix.mapRect(child->boundingRect() | child->childrenBoundingRect());
    }
    QPointF zero = childRect.center();
    //log.info(QString("center [%1 %2]").arg(zero.x()).arg(zero.y()));
    //FIXME does not work
    //views().first()->centerOn(zero);
    update();
}

}//namespace
