/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "ProjectImpl.h"

#include <core_api/ServiceTypes.h>
#include <core_api/ObjectViewModel.h>
#include <core_api/DocumentModel.h>

#include <core_api/GHints.h>

#include <core_api/Log.h>

namespace GB2 {

/* TRANSLATOR GB2::ProjectImpl */

static LogCategory log(ULOG_CAT_CORE_SERVICES);



ProjectImpl::ProjectImpl(const QString& _name, const QString& _url, const QList<Document*>& _docs, const QList<GObjectViewState*>& _states)
: name(_name), url(_url) 
{
    mainThreadModel = true;
    foreach(Document* doc, _docs) {
        addDocument(doc);
    }
    foreach(GObjectViewState* state, _states) {
        addGObjectViewState(state);
    }
    setModified(false);
}

ProjectImpl::~ProjectImpl() {
	//delete all docs
	while (!docs.isEmpty()) {
		Document* d = docs.takeLast();
		delete d;
	}

	//delete all views
	while (!objectViews.isEmpty()) {
		GObjectViewState* s = objectViews.takeLast();
		delete s;
	}
}

void ProjectImpl::makeClean() {
	if (!isTreeItemModified()) {
		return;
	}
	setModified(false);
	foreach(Document* d, docs) {
		d->makeClean();
	}
	assert(!isTreeItemModified());
}

void ProjectImpl::setProjectName(const QString& newName) {
	if (name==newName) {
		return;
	}
	setModified(true);
	name = newName;
	emit si_projectRenamed(this);
}

void ProjectImpl::setProjectURL(const QString& newURL) {
    if (url == newURL) {
        return;
    }
    
    log.details(tr("Project URL is set to %1\n").arg(newURL));

    setModified(true);
    QString oldURL = url;
    url = newURL;
    emit si_projectURLChanged(oldURL);
}

Document* ProjectImpl::findDocumentByURL(const QString & url) const {
	foreach(Document* d, docs) {
		if (d->getURLString() == url) {
			return d;
		}
	}
	return NULL;
}

void ProjectImpl::addDocument(Document* d) {
	log.details(tr("add_document_url_%1").arg(d->getURLString()));

	assert(findDocumentByURL(d->getURL())==NULL);
	setParentStateLockItem_static(d, this);

    d->setGHints(new ModTrackHints(this, d->getGHintsMap(), true));

	docs.push_back(d);
	emit si_documentAdded(d);
	setModified(true);
}

void ProjectImpl::removeDocument(Document* d, bool autodelete) {
	log.details(tr("remove_documenturl_%1").arg(d->getURLString()));

	setParentStateLockItem_static(d, NULL);
	docs.removeOne(d);

    d->setGHints(new GHintsDefaultImpl(d->getGHints()->getMap()));

	emit si_documentRemoved(d);
	if (autodelete) {
		delete d;
	}
    setModified(true);
} 


void ProjectImpl::sl_onStateModified(GObjectViewState*) {
	setModified(true);
}

void ProjectImpl::addState(GObjectViewState* s) {
	assert(!objectViews.contains(s));
	connect(s, SIGNAL(si_stateModified(GObjectViewState*)), SLOT(sl_onStateModified(GObjectViewState*)));
	objectViews.append(s);
	setModified(true);
}

void ProjectImpl::addGObjectViewState(GObjectViewState* s) {
	assert(GObjectViewUtils::findStateInList(s->getViewName(), s->getStateName(), objectViews) == NULL);
	addState(s);
	emit si_objectViewStateAdded(s);
}

void ProjectImpl::removeGObjectViewState(GObjectViewState* s) {
	int i = objectViews.removeAll(s);
    Q_UNUSED(i);
    assert(i == 1);
	emit si_objectViewStateRemoved(s);
	setModified(true);
}

}
